"""Chess source helpers"""

from dlt.common.typing import StrAny
from dlt.sources.helpers import requests

from .settings import OFFICIAL_CHESS_API_URL


def get_url_with_retry(url: str) -> StrAny:
    r = requests.get(url)
    return r.json()  # type: ignore


def get_path_with_retry(path: str) -> StrAny:
    return get_url_with_retry(f"{OFFICIAL_CHESS_API_URL}{path}")


def validate_month_string(string: str) -> None:
    """Validates that the string is in YYYY/MM format"""
    if string and string[4] != "/":
        raise ValueError(string)



import json
from datetime import date, datetime
from typing import Any, Iterator, Optional

import dlt
import proto  # type: ignore
from dlt.common.exceptions import MissingDependencyException
from dlt.common.typing import TDataItem
from dlt.sources import DltResource
from flatten_json import flatten  # type: ignore
from googleapiclient.discovery import Resource  # type: ignore

from . import field
from .metrics import dlt_metrics_schema
from .predicates import date_predicate
from .reports import BUILTIN_REPORTS, Report

try:
    from google.ads.googleads.client import GoogleAdsClient  # type: ignore
except ImportError:
    raise MissingDependencyException("Requests-OAuthlib", ["google-ads"])


@dlt.source
def google_ads(
    client: GoogleAdsClient,
    customer_id: str,
    report_spec: Optional[str] = None,
    start_date: Optional[datetime] = None,
    end_date: Optional[datetime] = None,
) -> Iterator[DltResource]:
    date_range = dlt.sources.incremental(
        "segments_date",
        initial_value=start_date.date(),  # type: ignore
        end_value=end_date.date() if end_date is not None else None,  # type: ignore
        range_start="closed",
        range_end="closed",
    )
    if report_spec is not None:
        custom_report = Report().from_spec(report_spec)
        yield dlt.resource(
            daily_report,
            name="daily_report",
            write_disposition="merge",
            primary_key=custom_report.primary_keys(),
            columns=dlt_metrics_schema(custom_report.metrics),
        )(client, customer_id, custom_report, date_range)

    for report_name, report in BUILTIN_REPORTS.items():
        yield dlt.resource(
            daily_report,
            name=report_name,
            write_disposition="merge",
            primary_key=report.primary_keys(),
            columns=dlt_metrics_schema(report.metrics),
        )(client, customer_id, report, date_range)


def daily_report(
    client: Resource,
    customer_id: str,
    report: Report,
    date: dlt.sources.incremental[date],
) -> Iterator[TDataItem]:
    ga_service = client.get_service("GoogleAdsService")
    fields = report.dimensions + report.metrics + report.segments
    criteria = date_predicate("segments.date", date.last_value, date.end_value)  # type:ignore
    query = f"""
        SELECT
            {", ".join(fields)}
        FROM
            {report.resource}
        WHERE
            {criteria}
    """
    if report.unfilterable is True:
        i = query.index("WHERE", 0)
        query = query[:i]

    allowed_keys = set([field.to_column(k) for k in fields])
    stream = ga_service.search_stream(customer_id=customer_id, query=query)
    for batch in stream:
        for row in batch.results:
            data = flatten(merge_lists(to_dict(row)))
            if "segments_date" in data:
                data["segments_date"] = datetime.strptime(
                    data["segments_date"], "%Y-%m-%d"
                ).date()
            yield {k: v for k, v in data.items() if k in allowed_keys}


def to_dict(item: Any) -> TDataItem:
    """
    Processes a batch result (page of results per dimension) accordingly
    :param batch:
    :return:
    """
    return json.loads(
        proto.Message.to_json(
            item,
            preserving_proto_field_name=True,
            use_integers_for_enums=False,
            including_default_value_fields=False,
        )
    )


def merge_lists(item: dict) -> dict:
    replacements = {}
    for k, v in item.get("metrics", {}).items():
        if isinstance(v, list):
            replacements[k] = ",".join(v)
    if len(replacements) == 0:
        return item
    item["metrics"].update(replacements)
    return item



"""
This is a module that provides a DLT source to retrieve data from multiple endpoints of the HubSpot API using a specified API key. The retrieved data is returned as a tuple of Dlt resources, one for each endpoint.

The source retrieves data from the following endpoints:
- CRM Companies
- CRM Contacts
- CRM Deals
- CRM Tickets
- CRM Products
- CRM Quotes
- Web Analytics Events

For each endpoint, a resource and transformer function are defined to retrieve data and transform it to a common format.
The resource functions yield the raw data retrieved from the API, while the transformer functions are used to retrieve
additional information from the Web Analytics Events endpoint.

The source also supports enabling Web Analytics Events for each endpoint by setting the corresponding enable flag to True.

Example:
To retrieve data from all endpoints, use the following code:

python

>>> resources = hubspot(api_key="your_api_key")
"""

from typing import Any, Dict, Iterator, List, Literal, Sequence
from urllib.parse import quote

import dlt
from dlt.common import pendulum
from dlt.common.typing import TDataItems
from dlt.sources import DltResource

from .helpers import _get_property_names, fetch_data, fetch_property_history
from .settings import (
    ALL,
    CRM_OBJECT_ENDPOINTS,
    DEFAULT_COMPANY_PROPS,
    DEFAULT_CONTACT_PROPS,
    DEFAULT_DEAL_PROPS,
    DEFAULT_PRODUCT_PROPS,
    DEFAULT_QUOTE_PROPS,
    DEFAULT_TICKET_PROPS,
    OBJECT_TYPE_PLURAL,
    STARTDATE,
    WEB_ANALYTICS_EVENTS_ENDPOINT,
)

THubspotObjectType = Literal["company", "contact", "deal", "ticket", "product", "quote"]


@dlt.source(name="hubspot", max_table_nesting=0)
def hubspot(
    api_key: str = dlt.secrets.value,
    include_history: bool = False,
    include_custom_props: bool = True,
) -> Sequence[DltResource]:
    """
    A DLT source that retrieves data from the HubSpot API using the
    specified API key.

    This function retrieves data for several HubSpot API endpoints,
    including companies, contacts, deals, tickets, products and web
    analytics events. It returns a tuple of Dlt resources, one for
    each endpoint.

    Args:
        api_key (Optional[str]):
            The API key used to authenticate with the HubSpot API. Defaults
            to dlt.secrets.value.
        include_history (Optional[bool]):
            Whether to load history of property changes along with entities.
            The history entries are loaded to separate tables.

    Returns:
        Sequence[DltResource]: Dlt resources, one for each HubSpot API endpoint.

    Notes:
        This function uses the `fetch_data` function to retrieve data from the
        HubSpot CRM API. The API key is passed to `fetch_data` as the
        `api_key` argument.
    """

    @dlt.resource(name="companies", write_disposition="replace")
    def companies(
        api_key: str = api_key,
        include_history: bool = include_history,
        props: Sequence[str] = DEFAULT_COMPANY_PROPS,
        include_custom_props: bool = include_custom_props,
    ) -> Iterator[TDataItems]:
        """Hubspot companies resource"""
        yield from crm_objects(
            "company",
            api_key,
            include_history=include_history,
            props=props,
            include_custom_props=include_custom_props,
        )

    @dlt.resource(name="contacts", write_disposition="replace")
    def contacts(
        api_key: str = api_key,
        include_history: bool = include_history,
        props: Sequence[str] = DEFAULT_CONTACT_PROPS,
        include_custom_props: bool = include_custom_props,
    ) -> Iterator[TDataItems]:
        """Hubspot contacts resource"""
        yield from crm_objects(
            "contact",
            api_key,
            include_history,
            props,
            include_custom_props,
        )

    @dlt.resource(name="deals", write_disposition="replace")
    def deals(
        api_key: str = api_key,
        include_history: bool = include_history,
        props: Sequence[str] = DEFAULT_DEAL_PROPS,
        include_custom_props: bool = include_custom_props,
    ) -> Iterator[TDataItems]:
        """Hubspot deals resource"""
        yield from crm_objects(
            "deal",
            api_key,
            include_history,
            props,
            include_custom_props,
        )

    @dlt.resource(name="tickets", write_disposition="replace")
    def tickets(
        api_key: str = api_key,
        include_history: bool = include_history,
        props: Sequence[str] = DEFAULT_TICKET_PROPS,
        include_custom_props: bool = include_custom_props,
    ) -> Iterator[TDataItems]:
        """Hubspot tickets resource"""
        yield from crm_objects(
            "ticket",
            api_key,
            include_history,
            props,
            include_custom_props,
        )

    @dlt.resource(name="products", write_disposition="replace")
    def products(
        api_key: str = api_key,
        include_history: bool = include_history,
        props: Sequence[str] = DEFAULT_PRODUCT_PROPS,
        include_custom_props: bool = include_custom_props,
    ) -> Iterator[TDataItems]:
        """Hubspot products resource"""
        yield from crm_objects(
            "product",
            api_key,
            include_history,
            props,
            include_custom_props,
        )

    @dlt.resource(name="quotes", write_disposition="replace")
    def quotes(
        api_key: str = api_key,
        include_history: bool = include_history,
        props: Sequence[str] = DEFAULT_QUOTE_PROPS,
        include_custom_props: bool = include_custom_props,
    ) -> Iterator[TDataItems]:
        """Hubspot quotes resource"""
        yield from crm_objects(
            "quote",
            api_key,
            include_history,
            props,
            include_custom_props,
        )

    return companies, contacts, deals, tickets, products, quotes


def crm_objects(
    object_type: str,
    api_key: str = dlt.secrets.value,
    include_history: bool = False,
    props: Sequence[str] = None,
    include_custom_props: bool = True,
) -> Iterator[TDataItems]:
    """Building blocks for CRM resources."""
    if props == ALL:
        props = list(_get_property_names(api_key, object_type))

    if include_custom_props:
        all_props = _get_property_names(api_key, object_type)
        custom_props = [prop for prop in all_props if not prop.startswith("hs_")]
        props = props + custom_props  # type: ignore

    props = ",".join(sorted(list(set(props))))


    params = {"properties": props, "limit": 100}

    yield from fetch_data(CRM_OBJECT_ENDPOINTS[object_type], api_key, params=params)
    if include_history:
        # Get history separately, as requesting both all properties and history together
        # is likely to hit hubspot's URL length limit
        for history_entries in fetch_property_history(
            CRM_OBJECT_ENDPOINTS[object_type],
            api_key,
            props,
        ):
            yield dlt.mark.with_table_name(
                history_entries,
                OBJECT_TYPE_PLURAL[object_type] + "_property_history",
            )


@dlt.resource
def hubspot_events_for_objects(
    object_type: THubspotObjectType,
    object_ids: List[str],
    api_key: str = dlt.secrets.value,
    start_date: pendulum.DateTime = STARTDATE,
) -> DltResource:
    """
    A standalone DLT resources that retrieves web analytics events from the HubSpot API for a particular object type and list of object ids.

    Args:
        object_type(THubspotObjectType, required): One of the hubspot object types see definition of THubspotObjectType literal
        object_ids: (List[THubspotObjectType], required): List of object ids to track events
        api_key (str, optional): The API key used to authenticate with the HubSpot API. Defaults to dlt.secrets.value.
        start_date (datetime, optional): The initial date time from which start getting events, default to STARTDATE

    Returns:
        incremental dlt resource to track events for objects from the list
    """

    end_date = pendulum.now().isoformat()
    name = object_type + "_events"

    def get_web_analytics_events(
        occurred_at: dlt.sources.incremental[str],
    ) -> Iterator[List[Dict[str, Any]]]:
        """
        A helper function that retrieves web analytics events for a given object type from the HubSpot API.

        Args:
            object_type (str): The type of object for which to retrieve web analytics events.

        Yields:
            dict: A dictionary representing a web analytics event.
        """
        for object_id in object_ids:
            yield from fetch_data(
                WEB_ANALYTICS_EVENTS_ENDPOINT.format(
                    objectType=object_type,
                    objectId=object_id,
                    occurredAfter=quote(occurred_at.last_value),
                    occurredBefore=quote(end_date),
                ),
                api_key=api_key,
            )

    return dlt.resource(
        get_web_analytics_events,
        name=name,
        primary_key="id",
        write_disposition="append",
        selected=True,
        table_name=lambda e: name + "_" + str(e["eventType"]),
    )(
        dlt.sources.incremental(
            "occurredAt",
            initial_value=start_date.isoformat(),
            range_end="closed",
            range_start="closed",
        )
    )



import datetime
from typing import Optional


def isotime(dt: Optional[datetime.datetime]) -> Optional[str]:
    """
    Converts a datetime object to an iso 8601 formatted string.
    """
    if dt is None:
        return None
    return dt.isoformat()



from datetime import timedelta
from typing import Iterator

import dlt
import pandas as pd  # type: ignore[import-untyped]
import pendulum
import requests
from dlt.sources import DltResource
from dlt.sources.helpers.requests import Client
from pendulum.date import Date


@dlt.source(max_table_nesting=0)
def applovin_max_source(
    start_date: Date,
    applications: list[str],
    api_key: str,
    end_date: Date | None,
) -> DltResource:
    @dlt.resource(
        name="user_ad_revenue",
        write_disposition="merge",
        merge_key="partition_date",
        columns={
            "partition_date": {"data_type": "date", "partition": True},
        },
    )
    def fetch_ad_revenue_report(
        dateTime=(
            dlt.sources.incremental(
                "partition_date",
                initial_value=start_date,
                end_value=end_date,
                range_start="closed",
                range_end="closed",
            )
        ),
    ) -> Iterator[dict]:
        url = "https://r.applovin.com/max/userAdRevenueReport"
        start_date = dateTime.last_value

        if dateTime.end_value is None:
            end_date = (pendulum.yesterday("UTC")).date()
        else:
            end_date = dateTime.end_value

        client = create_client()
        platforms = ["ios", "android", "fireos"]

        for app in applications:
            current_date = start_date
            while current_date <= end_date:
                for platform in platforms:
                    df = get_data(
                        url=url,
                        current_date=current_date,
                        application=app,
                        api_key=api_key,
                        client=client,
                        platform=platform,
                    )
                    if df is not None:
                        yield df
                current_date = current_date + timedelta(days=1)

    return fetch_ad_revenue_report


def create_client() -> requests.Session:
    return Client(
        request_timeout=10.0,
        raise_for_status=False,
        retry_condition=retry_on_limit,
        request_max_attempts=12,
    ).session


def retry_on_limit(
    response: requests.Response | None, exception: BaseException | None
) -> bool:
    if response is None:
        return False
    return response.status_code == 429


def get_data(
    url: str,
    current_date: Date,
    application: str,
    api_key: str,
    platform: str,
    client: requests.Session,
):
    params = {
        "api_key": api_key,
        "date": current_date.isoformat(),
        "platform": platform,
        "application": application,
        "aggregated": "false",
    }

    response = client.get(url=url, params=params)

    if response.status_code != 200:
        if response.status_code == 404:
            if "No Mediation App Id found for platform" in response.text:
                return None
        error_message = f"AppLovin MAX API error (status {response.status_code}): {response.text}"
        raise requests.HTTPError(error_message)
    
    response_url = response.json().get("ad_revenue_report_url")
    df = pd.read_csv(response_url)
    df["Date"] = pd.to_datetime(df["Date"])
    df["partition_date"] = df["Date"].dt.date
    return df


import base64
import csv
import json
import os
import shutil
import tempfile
from urllib.parse import parse_qs, quote, urlparse

import dlt
from dlt.common.configuration.specs import AwsCredentials
from dlt.destinations.impl.clickhouse.configuration import (
    ClickHouseCredentials,
)

from ingestr.src.loader import load_dlt_file


class GenericSqlDestination:
    def dlt_run_params(self, uri: str, table: str, **kwargs) -> dict:
        table_fields = table.split(".")
        if len(table_fields) != 2:
            raise ValueError("Table name must be in the format <schema>.<table>")

        res = {
            "dataset_name": table_fields[-2],
            "table_name": table_fields[-1],
        }

        return res

    def post_load(self):
        pass


class BigQueryDestination:
    def dlt_dest(self, uri: str, **kwargs):
        source_fields = urlparse(uri)
        source_params = parse_qs(source_fields.query)

        cred_path = source_params.get("credentials_path")
        credentials_base64 = source_params.get("credentials_base64")
        if not cred_path and not credentials_base64:
            raise ValueError(
                "credentials_path or credentials_base64 is required to connect BigQuery"
            )

        location = None
        if source_params.get("location"):
            loc_params = source_params.get("location", [])
            if len(loc_params) > 1:
                raise ValueError("Only one location is allowed")
            location = loc_params[0]

        credentials = {}
        if cred_path:
            with open(cred_path[0], "r") as f:
                credentials = json.load(f)
        elif credentials_base64:
            credentials = json.loads(
                base64.b64decode(credentials_base64[0]).decode("utf-8")
            )

        project_id = None
        if source_fields.hostname:
            project_id = source_fields.hostname

        return dlt.destinations.bigquery(
            credentials=credentials,  # type: ignore
            location=location,
            project_id=project_id,
            **kwargs,
        )

    def dlt_run_params(self, uri: str, table: str, **kwargs) -> dict:
        table_fields = table.split(".")
        if len(table_fields) != 2 and len(table_fields) != 3:
            raise ValueError(
                "Table name must be in the format <dataset>.<table> or <project>.<dataset>.<table>"
            )

        res = {
            "dataset_name": table_fields[-2],
            "table_name": table_fields[-1],
        }

        return res

    def post_load(self):
        pass


class PostgresDestination(GenericSqlDestination):
    def dlt_dest(self, uri: str, **kwargs):
        return dlt.destinations.postgres(credentials=uri, **kwargs)


class SnowflakeDestination(GenericSqlDestination):
    def dlt_dest(self, uri: str, **kwargs):
        return dlt.destinations.snowflake(credentials=uri, **kwargs)


class RedshiftDestination(GenericSqlDestination):
    def dlt_dest(self, uri: str, **kwargs):
        return dlt.destinations.redshift(
            credentials=uri.replace("redshift://", "postgresql://"), **kwargs
        )


class DuckDBDestination(GenericSqlDestination):
    def dlt_dest(self, uri: str, **kwargs):
        return dlt.destinations.duckdb(uri, **kwargs)


class MsSQLDestination(GenericSqlDestination):
    def dlt_dest(self, uri: str, **kwargs):
        return dlt.destinations.mssql(credentials=uri, **kwargs)


class DatabricksDestination(GenericSqlDestination):
    def dlt_dest(self, uri: str, **kwargs):
        return dlt.destinations.databricks(credentials=uri, **kwargs)


class SynapseDestination(GenericSqlDestination):
    def dlt_dest(self, uri: str, **kwargs):
        return dlt.destinations.synapse(credentials=uri, **kwargs)


class CustomCsvDestination(dlt.destinations.filesystem):
    pass


class CsvDestination(GenericSqlDestination):
    temp_path: str
    actual_path: str
    uri: str
    dataset_name: str
    table_name: str

    def dlt_run_params(self, uri: str, table: str, **kwargs) -> dict:
        table_fields = table.split(".")
        if len(table_fields) != 2:
            raise ValueError("Table name must be in the format <schema>.<table>")

        res = {
            "dataset_name": table_fields[-2],
            "table_name": table_fields[-1],
        }

        self.dataset_name = res["dataset_name"]
        self.table_name = res["table_name"]
        self.uri = uri

        return res

    def dlt_dest(self, uri: str, **kwargs):
        if uri.startswith("csv://"):
            uri = uri.replace("csv://", "file://")

        temp_path = tempfile.mkdtemp()
        self.actual_path = uri
        self.temp_path = temp_path
        return CustomCsvDestination(bucket_url=f"file://{temp_path}", **kwargs)

    # I dislike this implementation quite a bit since it ties the implementation to some internal details on how dlt works
    # I would prefer a custom destination that allows me to do this easily but dlt seems to have a lot of internal details that are not documented
    # I tried to make it work with a nicer destination implementation but I couldn't, so I decided to go with this hack to experiment
    # if anyone has a better idea on how to do this, I am open to contributions or suggestions
    def post_load(self):
        def find_first_file(path):
            for entry in os.listdir(path):
                full_path = os.path.join(path, entry)
                if os.path.isfile(full_path):
                    return full_path

            return None

        def filter_keys(dictionary):
            return {
                key: value
                for key, value in dictionary.items()
                if not key.startswith("_dlt_")
            }

        first_file_path = find_first_file(
            f"{self.temp_path}/{self.dataset_name}/{self.table_name}"
        )

        output_path = self.uri.split("://")[1]
        if output_path.count("/") > 1:
            os.makedirs(os.path.dirname(output_path), exist_ok=True)

        with open(output_path, "w", newline="") as csv_file:
            csv_writer = None
            for row in load_dlt_file(first_file_path):
                row = filter_keys(row)
                if csv_writer is None:
                    csv_writer = csv.DictWriter(csv_file, fieldnames=row.keys())
                    csv_writer.writeheader()

                csv_writer.writerow(row)
        shutil.rmtree(self.temp_path)


class AthenaDestination:
    def dlt_dest(self, uri: str, **kwargs):
        encoded_uri = quote(uri, safe=":/?&=")
        source_fields = urlparse(encoded_uri)
        source_params = parse_qs(source_fields.query)

        bucket = source_params.get("bucket", [None])[0]
        if not bucket:
            raise ValueError("A bucket is required to connect to Athena.")

        if not bucket.startswith("s3://"):
            bucket = f"s3://{bucket}"

        query_result_path = source_params.get("query_results_path", [None])[0]
        if query_result_path:
            if not query_result_path.startswith("s3://"):
                query_result_path = f"s3://{query_result_path}"
        else:
            query_result_path = bucket

        access_key_id = source_params.get("access_key_id", [None])[0]
        if not access_key_id:
            raise ValueError("The AWS access_key_id is required to connect to Athena.")

        secret_access_key = source_params.get("secret_access_key", [None])[0]
        if not secret_access_key:
            raise ValueError("The AWS secret_access_key is required to connect Athena")

        work_group = source_params.get("workgroup", [None])[0]

        region_name = source_params.get("region_name", [None])[0]
        if not region_name:
            raise ValueError("The region_name is required to connect to Athena.")

        os.environ["DESTINATION__BUCKET_URL"] = bucket
        os.environ["DESTINATION__CREDENTIALS__AWS_ACCESS_KEY_ID"] = access_key_id
        os.environ["DESTINATION__CREDENTIALS__AWS_SECRET_ACCESS_KEY"] = (
            secret_access_key
        )

        credentials = AwsCredentials(
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
            region_name=region_name,
        )
        return dlt.destinations.athena(
            query_result_bucket=query_result_path,
            athena_work_group=work_group,
            credentials=credentials,
            destination_name=bucket,
        )

    def dlt_run_params(self, uri: str, table: str, **kwargs) -> dict:
        table_fields = table.split(".")
        if len(table_fields) != 2:
            raise ValueError("Table name must be in the format <schema>.<table>")
        return {
            "table_format": "iceberg",
            "dataset_name": table_fields[-2],
            "table_name": table_fields[-1],
        }

    def post_load(self):
        pass


class ClickhouseDestination:
    def dlt_dest(self, uri: str, **kwargs):
        parsed_uri = urlparse(uri)

        if "dest_table" in kwargs:
            table = kwargs["dest_table"]
            database = table.split(".")[0]
        else:
            database = parsed_uri.path.lstrip("/")

        username = parsed_uri.username
        if not username:
            raise ValueError(
                "A username is required to connect to the ClickHouse database."
            )

        password = parsed_uri.password
        if not password:
            raise ValueError(
                "A password is required to authenticate with the ClickHouse database."
            )

        host = parsed_uri.hostname
        if not host:
            raise ValueError(
                "The hostname or IP address of the ClickHouse server is required to establish a connection."
            )

        port = parsed_uri.port
        if not port:
            raise ValueError(
                "The TCP port of the ClickHouse server is required to establish a connection."
            )

        query_params = parse_qs(parsed_uri.query)
        secure = int(query_params["secure"][0]) if "secure" in query_params else 1

        http_port = (
            int(query_params["http_port"][0])
            if "http_port" in query_params
            else 8443
            if secure == 1
            else 8123
        )

        if secure not in (0, 1):
            raise ValueError(
                "Invalid value for secure. Set to `1` for a secure HTTPS connection or `0` for a non-secure HTTP connection."
            )

        credentials = ClickHouseCredentials(
            {
                "host": host,
                "port": port,
                "username": username,
                "password": password,
                "database": database,
                "http_port": http_port,
                "secure": secure,
            }
        )
        return dlt.destinations.clickhouse(credentials=credentials)

    def dlt_run_params(self, uri: str, table: str, **kwargs) -> dict:
        table_fields = table.split(".")
        if len(table_fields) != 2:
            raise ValueError("Table name must be in the format <schema>.<table>")
        return {
            "table_name": table_fields[-1],
        }

    def post_load(self):
        pass



from dlt.common.libs.sql_alchemy import Table


def cast_set_to_list(row):
    # this handles just the sqlalchemy backend for now
    if isinstance(row, dict):
        for key in row.keys():
            if isinstance(row[key], set):
                row[key] = list(row[key])
    return row


def table_adapter_exclude_columns(cols: list[str]):
    def excluder(table: Table):
        cols_to_remove = [col for col in table._columns if col.name in cols]  # type: ignore
        for col in cols_to_remove:
            table._columns.remove(col)  # type: ignore

    return excluder



from typing import Optional

import pendulum
import requests
from dlt.sources.helpers.requests import Client
from requests.exceptions import HTTPError

DEFAULT_DIMENSIONS = ["campaign", "day", "app", "store_type", "channel", "country"]

DEFAULT_METRICS = [
    "network_cost",
    "all_revenue_total_d0",
    "ad_revenue_total_d0",
    "revenue_total_d0",
    "all_revenue_total_d1",
    "ad_revenue_total_d1",
    "revenue_total_d1",
    "all_revenue_total_d3",
    "ad_revenue_total_d3",
    "revenue_total_d3",
    "all_revenue_total_d7",
    "ad_revenue_total_d7",
    "revenue_total_d7",
    "all_revenue_total_d14",
    "ad_revenue_total_d14",
    "revenue_total_d14",
    "all_revenue_total_d21",
]


def retry_on_limit(response: requests.Response, exception: BaseException) -> bool:
    return response.status_code == 429


class AdjustAPI:
    def __init__(self, api_key):
        self.api_key = api_key
        self.request_client = Client(
            request_timeout=8.0,
            raise_for_status=False,
            retry_condition=retry_on_limit,
            request_max_attempts=12,
            request_backoff_factor=2,
        ).session

    def fetch_report_data(
        self,
        start_date: pendulum.DateTime,
        end_date: pendulum.DateTime,
        dimensions=DEFAULT_DIMENSIONS,
        metrics=DEFAULT_METRICS,
        filters: Optional[dict] = None,
    ):
        headers = {"Authorization": f"Bearer {self.api_key}"}
        params = {}

        if filters:
            for key, value in filters.items():
                if isinstance(value, list):
                    params[key] = ",".join(value)
                else:
                    params[key] = value

        params["date_period"] = (
            f"{start_date.format('YYYY-MM-DD')}:{end_date.format('YYYY-MM-DD')}"
        )
        params["dimensions"] = ",".join(dimensions)
        params["metrics"] = ",".join(metrics)

        if start_date > end_date:
            raise ValueError(
                f"Invalid date range: Start date ({start_date}) must be earlier than end date ({end_date})."
            )

        response = self.request_client.get(
            "https://automate.adjust.com/reports-service/report",
            headers=headers,
            params=params,
        )
        if response.status_code == 200:
            result = response.json()
            items = result.get("rows", [])
            yield items
        else:
            raise HTTPError(f"Request failed with status code: {response.status_code}, {response.text}.")

    def fetch_events(self):
        headers = {"Authorization": f"Bearer {self.api_key}"}
        response = self.request_client.get(
            "https://automate.adjust.com/reports-service/events", headers=headers
        )
        if response.status_code == 200:
            result = response.json()
            yield result
        else:
            raise HTTPError(f"Request failed with status code: {response.status_code}, {response.text}.")


def parse_filters(filters_raw: str) -> dict:
    # Parse filter string like "key1=value1,key2=value2,value3,value4"
    filters = {}
    current_key = None

    for item in filters_raw.split(","):
        if "=" in item:
            # Start of a new key-value pair
            key, value = item.split("=")
            filters[key] = [value]  # Always start with a list
            current_key = key
        elif current_key is not None:
            # Additional value for the current key
            filters[current_key].append(item)

    # Convert single-item lists to simple values
    filters = {k: v[0] if len(v) == 1 else v for k, v in filters.items()}

    return filters



"""Stripe analytics source helpers"""

from typing import Any, Dict, Iterable, Optional, Union

import stripe
from dlt.common import pendulum
from dlt.common.typing import TDataItem
from pendulum import DateTime


def pagination(
    endpoint: str, start_date: Optional[Any] = None, end_date: Optional[Any] = None
) -> Iterable[TDataItem]:
    """
    Retrieves data from an endpoint with pagination.

    Args:
        endpoint (str): The endpoint to retrieve data from.
        start_date (Optional[Any]): An optional start date to limit the data retrieved. Defaults to None.
        end_date (Optional[Any]): An optional end date to limit the data retrieved. Defaults to None.

    Returns:
        Iterable[TDataItem]: Data items retrieved from the endpoint.
    """
    starting_after = None
    while True:
        response = stripe_get_data(
            endpoint,
            start_date=start_date,
            end_date=end_date,
            starting_after=starting_after,
        )

        if len(response["data"]) > 0:
            starting_after = response["data"][-1]["id"]
        yield response["data"]

        if not response["has_more"]:
            break


def transform_date(date: Union[str, DateTime, int]) -> int:
    if isinstance(date, str):
        date = pendulum.from_format(date, "%Y-%m-%dT%H:%M:%SZ")
    if isinstance(date, DateTime):
        # convert to unix timestamp
        date = int(date.timestamp())
    return date


def stripe_get_data(
    resource: str,
    start_date: Optional[Any] = None,
    end_date: Optional[Any] = None,
    **kwargs: Any,
) -> Dict[Any, Any]:
    if start_date:
        start_date = transform_date(start_date)
    if end_date:
        end_date = transform_date(end_date)

    if resource == "Subscription":
        kwargs.update({"status": "all"})

    resource_dict = getattr(stripe, resource).list(
        created={"gte": start_date, "lt": end_date}, limit=100, **kwargs
    )
    return dict(resource_dict)



from typing import Dict

from dlt.common.schema.typing import TColumnSchema
from dlt.destinations.adapters import athena_adapter, athena_partition
from dlt.sources import DltResource, DltSource

import ingestr.src.resource as resource


def apply_athena_hints(
    source: DltSource | DltResource,
    partition_column: str,
    additional_hints: Dict[str, TColumnSchema] = {},
) -> None:
    def _apply_partition_hint(resource: DltResource) -> None:

        columns = resource.columns if resource.columns else {}

        partition_hint = (
            columns.get(partition_column)  # type: ignore
            or additional_hints.get(partition_column)
        )

        athena_adapter(
            resource,
            athena_partition.day(partition_column)
            if partition_hint and partition_hint.get("data_type") in ("timestamp", "date")
            else partition_column,
        )

    resource.for_each(source, _apply_partition_hint)



"""Fetches Shopify Orders and Products."""

from typing import Any, Dict, Iterable, Optional  # noqa: F401

import dlt
from dlt.common import jsonpath as jp  # noqa: F401
from dlt.common import pendulum
from dlt.common.time import ensure_pendulum_datetime
from dlt.common.typing import TAnyDateTime, TDataItem
from dlt.sources import DltResource

from .helpers import ShopifyApi, ShopifyGraphQLApi, TOrderStatus
from .settings import (
    DEFAULT_API_VERSION,
    DEFAULT_ITEMS_PER_PAGE,
    DEFAULT_PARTNER_API_VERSION,  # noqa: F401
    FIRST_DAY_OF_MILLENNIUM,
)


@dlt.source(name="shopify", max_table_nesting=0)
def shopify_source(
    private_app_password: str = dlt.secrets.value,
    api_version: str = DEFAULT_API_VERSION,
    shop_url: str = dlt.config.value,
    start_date: TAnyDateTime = FIRST_DAY_OF_MILLENNIUM,
    end_date: Optional[TAnyDateTime] = None,
    created_at_min: TAnyDateTime = FIRST_DAY_OF_MILLENNIUM,
    items_per_page: int = DEFAULT_ITEMS_PER_PAGE,
    order_status: TOrderStatus = "any",
) -> Iterable[DltResource]:
    """
    The source for the Shopify pipeline. Available resources are products, orders, and customers.

    `start_time` argument can be used on its own or together with `end_time`. When both are provided
    data is limited to items updated in that time range.
    The range is "half-open", meaning elements equal and newer than `start_time` and elements older than `end_time` are included.
    All resources opt-in to use Airflow scheduler if run as Airflow task

    Args:
        private_app_password: The app password to the app on your shop.
        api_version: The API version to use (e.g. 2023-01).
        shop_url: The URL of your shop (e.g. https://my-shop.myshopify.com).
        items_per_page: The max number of items to fetch per page. Defaults to 250.
        start_date: Items updated on or after this date are imported. Defaults to 2000-01-01.
            If end date is not provided, this is used as the initial value for incremental loading and after the initial run, only new data will be retrieved.
            Accepts any `date`/`datetime` object or a date/datetime string in ISO 8601 format.
        end_time: The end time of the range for which to load data.
            Should be used together with `start_date` to limit the data to items updated in that time range.
            If end time is not provided, the incremental loading will be enabled and after initial run, only new data will be retrieved
        created_at_min: The minimum creation date of items to import. Items created on or after this date are loaded. Defaults to 2000-01-01.
        order_status: The order status to filter by. Can be 'open', 'closed', 'cancelled', or 'any'. Defaults to 'any'.

    Returns:
        Iterable[DltResource]: A list of DltResource objects representing the data resources.
    """

    # build client
    client = ShopifyApi(shop_url, private_app_password, api_version)

    start_date_obj = ensure_pendulum_datetime(start_date)
    end_date_obj = ensure_pendulum_datetime(end_date) if end_date else None
    created_at_min_obj = ensure_pendulum_datetime(created_at_min)

    # define resources
    @dlt.resource(
        primary_key="id",
        write_disposition="merge",
        columns={
            "body_html": {
                "data_type": "text",
                "nullable": True,
                "description": "A description of the product. Supports HTML formatting.",
            },
            "created_at": {
                "data_type": "timestamp",
                "nullable": False,
                "description": "The date and time (ISO 8601 format) when the product was created.",
            },
            "handle": {
                "data_type": "text",
                "nullable": False,
                "description": "A unique human-friendly string for the product. Automatically generated from the product's title. Used by the Liquid templating language to refer to objects.",
            },
            "id": {
                "data_type": "bigint",
                "nullable": False,
                "primary_key": True,
                "description": "An unsigned 64-bit integer that's used as a unique identifier for the product.",
            },
            "images": {
                "data_type": "json",
                "nullable": True,
                "description": "A list of product image objects, each one representing an image associated with the product.",
            },
            "options": {
                "data_type": "json",
                "nullable": True,
                "description": "The custom product properties. For example, Size, Color, and Material.",
            },
            "product_type": {
                "data_type": "text",
                "nullable": True,
                "description": "A categorization for the product used for filtering and searching products.",
            },
            "published_at": {
                "data_type": "timestamp",
                "nullable": True,
                "description": "The date and time (ISO 8601 format) when the product was published.",
            },
            "published_scope": {
                "data_type": "text",
                "nullable": True,
                "description": "Whether the product is published to the Point of Sale channel.",
            },
            "status": {
                "data_type": "text",
                "nullable": False,
                "description": "The status of the product.",
            },
            "tags": {
                "data_type": "text",
                "nullable": True,
                "description": "A string of comma-separated tags used for filtering and search.",
            },
            "template_suffix": {
                "data_type": "text",
                "nullable": True,
                "description": "The suffix of the Liquid template used for the product page.",
            },
            "title": {
                "data_type": "text",
                "nullable": False,
                "description": "The name of the product.",
            },
            "updated_at": {
                "data_type": "timestamp",
                "nullable": True,
                "description": "The date and time (ISO 8601 format) when the product was last modified.",
            },
            "variants": {
                "data_type": "json",
                "nullable": True,
                "description": "An array of product variants, each representing a different version of the product.",
            },
            "vendor": {
                "data_type": "text",
                "nullable": True,
                "description": "The name of the product's vendor.",
            },
        },
    )
    def products_legacy(
        updated_at: dlt.sources.incremental[
            pendulum.DateTime
        ] = dlt.sources.incremental(
            "updated_at",
            initial_value=start_date_obj,
            end_value=end_date_obj,
            allow_external_schedulers=True,
            range_end="closed",
            range_start="closed",
        ),
        created_at_min: pendulum.DateTime = created_at_min_obj,
        items_per_page: int = items_per_page,
    ) -> Iterable[TDataItem]:
        """
        The resource for products on your shop, supports incremental loading and pagination.

        Args:
            updated_at: The saved state of the last 'updated_at' value.

        Returns:
            Iterable[TDataItem]: A generator of products.
        """
        params = dict(
            updated_at_min=updated_at.last_value.isoformat(),
            limit=items_per_page,
            order="updated_at asc",
            created_at_min=created_at_min.isoformat(),
        )
        if updated_at.end_value is not None:
            params["updated_at_max"] = updated_at.end_value.isoformat()
        yield from client.get_pages("products", params)

    @dlt.resource(
        primary_key="id",
        write_disposition="merge",
        columns={
            "app_id": {
                "data_type": "bigint",
                "nullable": True,
                "description": "The ID of the app that created the order.",
            },
            "billing_address": {
                "data_type": "json",
                "nullable": True,
                "description": "The mailing address associated with the payment method.",
            },
            "browser_ip": {
                "data_type": "text",
                "nullable": True,
                "description": "The IP address of the browser used by the customer when they placed the order.",
            },
            "buyer_accepts_marketing": {
                "data_type": "bool",
                "nullable": True,
                "description": "Whether the customer consented to receive email updates from the shop.",
            },
            "cancel_reason": {
                "data_type": "text",
                "nullable": True,
                "description": "The reason why the order was canceled.",
            },
            "cancelled_at": {
                "data_type": "timestamp",
                "nullable": True,
                "description": "The date and time when the order was canceled.",
            },
            "cart_token": {
                "data_type": "text",
                "nullable": True,
                "description": "A unique value referencing the cart associated with the order.",
            },
            "checkout_token": {
                "data_type": "text",
                "nullable": True,
                "description": "A unique value referencing the checkout associated with the order.",
            },
            "client_details": {
                "data_type": "json",
                "nullable": True,
                "description": "Information about the browser the customer used when placing the order.",
            },
            "closed_at": {
                "data_type": "timestamp",
                "nullable": True,
                "description": "The date and time when the order was closed.",
            },
            "company": {
                "data_type": "json",
                "nullable": True,
                "description": "Information about the purchasing company for the order.",
            },
            "confirmation_number": {
                "data_type": "text",
                "nullable": True,
                "description": "A randomly generated identifier for the order.",
            },
            "confirmed": {
                "data_type": "bool",
                "nullable": True,
                "description": "Whether inventory has been reserved for the order.",
            },
            "created_at": {
                "data_type": "timestamp",
                "nullable": False,
                "description": "The autogenerated date and time when the order was created.",
            },
            "currency": {
                "data_type": "text",
                "nullable": False,
                "description": "The three-letter code (ISO 4217 format) for the shop currency.",
            },
            "current_total_additional_fees_set": {
                "data_type": "json",
                "nullable": True,
                "description": "The current total additional fees on the order in shop and presentment currencies.",
            },
            "current_total_discounts": {
                "data_type": "decimal",
                "nullable": True,
                "description": "The current total discounts on the order in the shop currency.",
            },
            "current_total_discounts_set": {
                "data_type": "json",
                "nullable": True,
                "description": "The current total discounts on the order in shop and presentment currencies.",
            },
            "current_total_duties_set": {
                "data_type": "json",
                "nullable": True,
                "description": "The current total duties charged on the order in shop and presentment currencies.",
            },
            "current_total_price": {
                "data_type": "decimal",
                "nullable": True,
                "description": "The current total price of the order in the shop currency.",
            },
            "current_total_price_set": {
                "data_type": "json",
                "nullable": True,
                "description": "The current total price of the order in shop and presentment currencies.",
            },
            "current_subtotal_price": {
                "data_type": "decimal",
                "nullable": True,
                "description": "The sum of prices for all line items after discounts and returns in the shop currency.",
            },
            "current_subtotal_price_set": {
                "data_type": "json",
                "nullable": True,
                "description": "The sum of the prices for all line items after discounts and returns in shop and presentment currencies.",
            },
            "current_total_tax": {
                "data_type": "decimal",
                "nullable": True,
                "description": "The sum of the prices for all tax lines applied to the order in the shop currency.",
            },
            "current_total_tax_set": {
                "data_type": "json",
                "nullable": True,
                "description": "The sum of the prices for all tax lines applied to the order in shop and presentment currencies.",
            },
            "customer": {
                "data_type": "json",
                "nullable": True,
                "description": "Information about the customer.",
            },
            "customer_locale": {
                "data_type": "text",
                "nullable": True,
                "description": "The two or three-letter language code, optionally followed by a region modifier.",
            },
            "discount_applications": {
                "data_type": "json",
                "nullable": True,
                "description": "An ordered list of stacked discount applications.",
            },
            "discount_codes": {
                "data_type": "json",
                "nullable": True,
                "description": "A list of discounts applied to the order.",
            },
            "email": {
                "data_type": "text",
                "nullable": True,
                "description": "The customer's email address.",
            },
            "estimated_taxes": {
                "data_type": "bool",
                "nullable": True,
                "description": "Whether taxes on the order are estimated.",
            },
            "financial_status": {
                "data_type": "text",
                "nullable": True,
                "description": "The status of payments associated with the order.",
            },
            "fulfillments": {
                "data_type": "json",
                "nullable": True,
                "description": "An array of fulfillments associated with the order.",
            },
            "fulfillment_status": {
                "data_type": "text",
                "nullable": True,
                "description": "The order's status in terms of fulfilled line items.",
            },
            "gateway": {
                "data_type": "text",
                "nullable": True,
                "description": "The payment gateway used.",
            },
            "id": {
                "data_type": "bigint",
                "nullable": False,
                "primary_key": True,
                "description": "The ID of the order, used for API purposes.",
            },
            "landing_site": {
                "data_type": "text",
                "nullable": True,
                "description": "The URL for the page where the buyer landed when they entered the shop.",
            },
            "line_items": {
                "data_type": "json",
                "nullable": True,
                "description": "A list of line item objects containing information about an item in the order.",
            },
            "location_id": {
                "data_type": "bigint",
                "nullable": True,
                "description": "The ID of one of the locations assigned to fulfill the order.",
            },
            "merchant_of_record_app_id": {
                "data_type": "bigint",
                "nullable": True,
                "description": "The application acting as Merchant of Record for the order.",
            },
            "name": {
                "data_type": "text",
                "nullable": True,
                "description": "The order name, generated by combining the order_number with the order prefix and suffix.",
            },
            "note": {
                "data_type": "text",
                "nullable": True,
                "description": "An optional note that a shop owner can attach to the order.",
            },
            "note_attributes": {
                "data_type": "json",
                "nullable": True,
                "description": "Extra information added to the order as key-value pairs.",
            },
            "number": {
                "data_type": "bigint",
                "nullable": True,
                "description": "The order's position in the shop's count of orders.",
            },
            "order_number": {
                "data_type": "bigint",
                "nullable": True,
                "description": "The order's position in the shop's count of orders, starting at 1001.",
            },
            "original_total_additional_fees_set": {
                "data_type": "json",
                "nullable": True,
                "description": "The original total additional fees on the order in shop and presentment currencies.",
            },
            "original_total_duties_set": {
                "data_type": "json",
                "nullable": True,
                "description": "The original total duties charged on the order in shop and presentment currencies.",
            },
            "payment_terms": {
                "data_type": "json",
                "nullable": True,
                "description": "The terms and conditions under which a payment should be processed.",
            },
            "payment_gateway_names": {
                "data_type": "json",
                "nullable": True,
                "description": "The list of payment gateways used for the order.",
            },
            "phone": {
                "data_type": "text",
                "nullable": True,
                "description": "The customer's phone number for receiving SMS notifications.",
            },
            "po_number": {
                "data_type": "text",
                "nullable": True,
                "description": "The purchase order number associated with the order.",
            },
            "presentment_currency": {
                "data_type": "text",
                "nullable": True,
                "description": "The presentment currency used to display prices to the customer.",
            },
            "processed_at": {
                "data_type": "timestamp",
                "nullable": True,
                "description": "The date and time when an order was processed.",
            },
            "referring_site": {
                "data_type": "text",
                "nullable": True,
                "description": "The website where the customer clicked a link to the shop.",
            },
            "refunds": {
                "data_type": "json",
                "nullable": True,
                "description": "A list of refunds applied to the order.",
            },
            "shipping_address": {
                "data_type": "json",
                "nullable": True,
                "description": "The mailing address where the order will be shipped.",
            },
            "shipping_lines": {
                "data_type": "json",
                "nullable": True,
                "description": "An array detailing the shipping methods used.",
            },
            "source_name": {
                "data_type": "text",
                "nullable": True,
                "description": "The source of the checkout.",
            },
            "source_identifier": {
                "data_type": "text",
                "nullable": True,
                "description": "The ID of the order placed on the originating platform.",
            },
            "source_url": {
                "data_type": "text",
                "nullable": True,
                "description": "A valid URL to the original order on the originating surface.",
            },
            "subtotal_price": {
                "data_type": "decimal",
                "nullable": True,
                "description": "The price of the order in the shop currency after discounts but before shipping, duties, taxes, and tips.",
            },
            "subtotal_price_set": {
                "data_type": "json",
                "nullable": True,
                "description": "The subtotal of the order in shop and presentment currencies after discounts but before shipping, duties, taxes, and tips.",
            },
            "tags": {
                "data_type": "text",
                "nullable": True,
                "description": "Tags attached to the order, formatted as a string of comma-separated values.",
            },
            "tax_lines": {
                "data_type": "json",
                "nullable": True,
                "description": "An array of tax line objects detailing taxes applied to the order.",
            },
            "taxes_included": {
                "data_type": "bool",
                "nullable": True,
                "description": "Whether taxes are included in the order subtotal.",
            },
            "test": {
                "data_type": "bool",
                "nullable": True,
                "description": "Whether this is a test order.",
            },
            "token": {
                "data_type": "text",
                "nullable": True,
                "description": "A unique value referencing the order.",
            },
            "total_discounts": {
                "data_type": "decimal",
                "nullable": True,
                "description": "The total discounts applied to the price of the order in the shop currency.",
            },
            "total_discounts_set": {
                "data_type": "json",
                "nullable": True,
                "description": "The total discounts applied to the price of the order in shop and presentment currencies.",
            },
            "total_line_items_price": {
                "data_type": "decimal",
                "nullable": True,
                "description": "The sum of all line item prices in the shop currency.",
            },
            "total_line_items_price_set": {
                "data_type": "json",
                "nullable": True,
                "description": "The total of all line item prices in shop and presentment currencies.",
            },
            "total_outstanding": {
                "data_type": "decimal",
                "nullable": True,
                "description": "The total outstanding amount of the order in the shop currency.",
            },
            "total_price": {
                "data_type": "decimal",
                "nullable": True,
                "description": "The sum of all line item prices, discounts, shipping, taxes, and tips in the shop currency.",
            },
            "total_price_set": {
                "data_type": "json",
                "nullable": True,
                "description": "The total price of the order in shop and presentment currencies.",
            },
            "total_shipping_price_set": {
                "data_type": "json",
                "nullable": True,
                "description": "The total shipping price of the order in shop and presentment currencies.",
            },
            "total_tax": {
                "data_type": "decimal",
                "nullable": True,
                "description": "The sum of the prices for all tax lines applied to the order in the shop currency.",
            },
            "total_tax_set": {
                "data_type": "json",
                "nullable": True,
                "description": "The sum of the prices for all tax lines applied to the order in shop and presentment currencies.",
            },
            "total_tip_received": {
                "data_type": "decimal",
                "nullable": True,
                "description": "The sum of all the tips in the order in the shop currency.",
            },
            "total_weight": {
                "data_type": "decimal",
                "nullable": True,
                "description": "The sum of all line item weights in grams.",
            },
            "updated_at": {
                "data_type": "timestamp",
                "nullable": True,
                "description": "The date and time when the order was last modified.",
            },
            "user_id": {
                "data_type": "bigint",
                "nullable": True,
                "description": "The ID of the user logged into Shopify POS who processed the order.",
            },
            "order_status_url": {
                "data_type": "text",
                "nullable": True,
                "description": "The URL pointing to the order status web page, if applicable.",
            },
        },
    )
    def orders(
        updated_at: dlt.sources.incremental[
            pendulum.DateTime
        ] = dlt.sources.incremental(
            "updated_at",
            initial_value=start_date_obj,
            end_value=end_date_obj,
            allow_external_schedulers=True,
            range_end="closed",
            range_start="closed",
        ),
        created_at_min: pendulum.DateTime = created_at_min_obj,
        items_per_page: int = items_per_page,
        status: TOrderStatus = order_status,
    ) -> Iterable[TDataItem]:
        """
        The resource for orders on your shop, supports incremental loading and pagination.

        Args:
            updated_at: The saved state of the last 'updated_at' value.

        Returns:
            Iterable[TDataItem]: A generator of orders.
        """
        params = dict(
            updated_at_min=updated_at.last_value.isoformat(),
            limit=items_per_page,
            status=status,
            order="updated_at asc",
            created_at_min=created_at_min.isoformat(),
        )
        if updated_at.end_value is not None:
            params["updated_at_max"] = updated_at.end_value.isoformat()
        yield from client.get_pages("orders", params)

    @dlt.resource(primary_key="id", write_disposition="merge")
    def customers(
        updated_at: dlt.sources.incremental[
            pendulum.DateTime
        ] = dlt.sources.incremental(
            "updated_at",
            initial_value=start_date_obj,
            end_value=end_date_obj,
            allow_external_schedulers=True,
            range_end="closed",
            range_start="closed",
        ),
        created_at_min: pendulum.DateTime = created_at_min_obj,
        items_per_page: int = items_per_page,
    ) -> Iterable[TDataItem]:
        """
        The resource for customers on your shop, supports incremental loading and pagination.

        Args:
            updated_at: The saved state of the last 'updated_at' value.

        Returns:
            Iterable[TDataItem]: A generator of customers.
        """
        params = dict(
            updated_at_min=updated_at.last_value.isoformat(),
            limit=items_per_page,
            order="updated_at asc",
            created_at_min=created_at_min.isoformat(),
        )
        if updated_at.end_value is not None:
            params["updated_at_max"] = updated_at.end_value.isoformat()
        yield from client.get_pages("customers", params)

    @dlt.resource(primary_key="id", write_disposition="append")
    def events(
        created_at: dlt.sources.incremental[
            pendulum.DateTime
        ] = dlt.sources.incremental(
            "created_at",
            initial_value=start_date_obj,
            end_value=end_date_obj,
            range_end="closed",
            range_start="closed",
        ),
        items_per_page: int = items_per_page,
    ) -> Iterable[TDataItem]:
        params = dict(
            created_at_min=created_at.last_value.isoformat(),
            limit=items_per_page,
            order="created_at asc",
        )
        yield from client.get_pages("events", params)

    @dlt.resource(primary_key="id", write_disposition="merge")
    def price_rules(
        updated_at: dlt.sources.incremental[
            pendulum.DateTime
        ] = dlt.sources.incremental(
            "updated_at",
            initial_value=start_date_obj,
            end_value=end_date_obj,
            range_end="closed",
            range_start="closed",
        ),
        items_per_page: int = items_per_page,
    ) -> Iterable[TDataItem]:
        params = dict(
            updated_at_min=updated_at.last_value.isoformat(),
            limit=items_per_page,
            order="updated_at asc",
        )
        yield from client.get_pages("price_rules", params)

    @dlt.resource(primary_key="id", write_disposition="merge")
    def transactions(
        since_id: dlt.sources.incremental[int] = dlt.sources.incremental(
            "id",
            initial_value=None,
        ),
        items_per_page: int = items_per_page,
    ) -> Iterable[TDataItem]:
        params = dict(
            limit=items_per_page,
        )
        if since_id.start_value is not None:
            params["since_id"] = since_id.start_value
        yield from client.get_pages("shopify_payments/balance/transactions", params)

    @dlt.resource(
        primary_key="currency",
        write_disposition={"disposition": "merge", "strategy": "scd2"},
    )
    def balance() -> Iterable[TDataItem]:
        yield from client.get_pages("shopify_payments/balance", {})

    @dlt.resource(primary_key="id", write_disposition="merge")
    def inventory_items(
        updated_at: dlt.sources.incremental[
            pendulum.DateTime
        ] = dlt.sources.incremental(
            "updatedAt",
            initial_value=start_date_obj,
            end_value=end_date_obj,
            allow_external_schedulers=True,
            range_end="closed",
            range_start="closed",
        ),
        items_per_page: int = items_per_page,
    ) -> Iterable[TDataItem]:
        client = ShopifyGraphQLApi(
            base_url=shop_url,
            access_token=private_app_password,
            api_version="2024-07",
        )

        query = """
            query inventoryItems($after: String, $query: String, $first: Int) {
            inventoryItems(after: $after, first: $first, query: $query) {
                edges {
                node {
                    id
                    countryCodeOfOrigin
                    createdAt
                    duplicateSkuCount
                    harmonizedSystemCode
                    inventoryHistoryUrl
                    legacyResourceId
                    measurement {
                    id
                    weight {
                        unit
                        value
                    }
                    }

                    provinceCodeOfOrigin
                    requiresShipping
                    sku
                    tracked
                    trackedEditable {
                    locked
                    reason
                    }
                    unitCost {
                    amount
                    currencyCode
                    }
                    updatedAt
                    variant {
                    id
                    availableForSale
                    barcode

                    compareAtPrice
                    createdAt
                    inventoryPolicy
                    inventoryQuantity
                    legacyResourceId

                    position
                    price
                    product {
                        id
                    }
                    requiresComponents

                    selectedOptions {
                        name
                        value
                    }
                    sellableOnlineQuantity

                    sellingPlanGroupsCount {
                        count
                        precision
                    }
                    sku

                    taxCode
                    taxable
                    title
                    updatedAt
                    }
                }
                }
                pageInfo {
                    endCursor
                }
            }
        }"""

        yield from client.get_graphql_pages(
            query,
            data_items_path="data.inventoryItems.edges[*].node",
            pagination_cursor_path="data.inventoryItems.pageInfo.endCursor",
            pagination_variable_name="after",
            variables={
                "query": f"updated_at:>'{updated_at.last_value.isoformat()}'",
                "first": items_per_page,
            },
        )

    @dlt.resource(primary_key="id", write_disposition="merge")
    def discounts(items_per_page: int = items_per_page) -> Iterable[TDataItem]:
        client = ShopifyGraphQLApi(
            base_url=shop_url,
            access_token=private_app_password,
            api_version="2024-07",
        )

        query = """
query discountNodes($after: String, $query: String, $first: Int)  {
  discountNodes(after: $after, first: $first, query: $query) {
    nodes {
      id
      discount {
        __typename
        ... on DiscountCodeApp {
          appDiscountType {
            app {
              id
            }
            functionId
            targetType
          }
          appliesOncePerCustomer
          asyncUsageCount
          combinesWith {
            orderDiscounts
            productDiscounts
            shippingDiscounts
          }
          codesCount {
            count
            precision
          }
          createdAt
          customerSelection {
            __typename
            ... on DiscountCustomerAll {
              allCustomers
            }
            ... on DiscountCustomerSegments {
              segments {
                creationDate
                id
                lastEditDate
                name
                query
              }
            }
            ... on DiscountCustomers {
              customers {
                id
              }
            }
          }
          discountClass
          discountId
          endsAt
          errorHistory {
            errorsFirstOccurredAt
            firstOccurredAt
            hasBeenSharedSinceLastError
            hasSharedRecentErrors
          }
          hasTimelineComment
          recurringCycleLimit
          shareableUrls {
            targetItemImage {
              id
              url
            }
            targetType
            title
            url
          }
          startsAt
          status
          title
          totalSales {
            amount
            currencyCode
          }
          updatedAt
          usageLimit
        }
        ... on DiscountCodeBasic {
          appliesOncePerCustomer
          asyncUsageCount
          codes: codes(first: 250) {
            nodes {
              id
              code
            }
          }
          codesCount {
            count
            precision
          }
          combinesWith {
            orderDiscounts
            productDiscounts
            shippingDiscounts
          }
          createdAt
          customerGets {
            appliesOnOneTimePurchase
            appliesOnSubscription
            items {
              __typename
              ... on AllDiscountItems {
                allItems
              }
              ... on DiscountCollections {
                collectionsFirst250: collections(first: 250) {
                  nodes {
                    id
                  }
                }
              }
              ... on DiscountProducts {
                productsFirst250: products(first: 250) {
                  nodes {
                    id
                  }
                }
                productVariantsFirst250: productVariants(first: 250) {
                  nodes {
                    id
                  }
                }
              }
            }
            value {
              __typename
              ... on DiscountAmount {
                amount {
                  amount
                  currencyCode
                }
                appliesOnEachItem
              }
              ... on DiscountOnQuantity {
                effect {
                  ... on DiscountAmount {
                    amount {
                      amount
                      currencyCode
                    }
                    appliesOnEachItem
                  }
                  ... on DiscountPercentage {
                    percentage
                  }
                }
                quantity {
                  quantity
                }
              }
              ... on DiscountPercentage {
                percentage
              }
            }
          }
          customerSelection {
            __typename
            ... on DiscountCustomerAll {
              allCustomers
            }
            ... on DiscountCustomerSegments {
              segments {
                creationDate
                id
                lastEditDate
                name
                query
              }
            }
            ... on DiscountCustomers {
              customers {
                id
              }
            }
          }
          discountClass
          endsAt
          hasTimelineComment
          minimumRequirement {
            __typename
            ... on DiscountMinimumQuantity {
              greaterThanOrEqualToQuantity
            }
            ... on DiscountMinimumSubtotal {
              greaterThanOrEqualToSubtotal {
                amount
                currencyCode
              }
            }
          }
          recurringCycleLimit
          shareableUrls {
            url
            title
          }
          shortSummary
          startsAt
          status
          summary
          title
          totalSales {
            amount
            currencyCode
          }
          updatedAt
          usageLimit
        }
        ... on DiscountCodeBxgy {
          appliesOncePerCustomer
          asyncUsageCount
          codesFirst50: codes(first: 50) {
            nodes {
              id
              code
            }
          }
          codesCount {
            count
            precision
          }
          combinesWith {
            orderDiscounts
            productDiscounts
            shippingDiscounts
          }
          createdAt
          customerBuys {
            items {
              __typename
              ... on AllDiscountItems {
                allItems
              }
              ... on DiscountCollections {
                collectionsFirst250: collections(first: 250) {
                  nodes {
                    id
                  }
                }
              }
              ... on DiscountProducts {
                productsFirst250: products(first: 250) {
                  nodes {
                    id
                  }
                }
                productVariantsFirst250: productVariants(first: 250) {
                  nodes {
                    id
                  }
                }
              }
            }
            value {
              __typename
              ... on DiscountPurchaseAmount {
                amount
              }
              ... on DiscountQuantity {
                quantity
              }
            }
          }
          customerGets {
            appliesOnOneTimePurchase
            appliesOnSubscription
            items {
              __typename
              ... on AllDiscountItems {
                allItems
              }
              ... on DiscountCollections {
                collectionsFirst250: collections(first: 250) {
                  nodes {
                    id
                  }
                }
              }
              ... on DiscountProducts {
                productsFirst250: products(first: 250) {
                  nodes {
                    id
                  }
                }
                productVariantsFirst250: productVariants(first: 250) {
                  nodes {
                    id
                  }
                }
              }
            }
            value {
              __typename
              ... on DiscountAmount {
                amount {
                  amount
                  currencyCode
                }
                appliesOnEachItem
              }
              ... on DiscountOnQuantity {
                effect {
                  __typename
                  ... on DiscountAmount {
                    amount {
                      amount
                      currencyCode
                    }
                    appliesOnEachItem
                  }
                  ... on DiscountPercentage {
                    percentage
                  }
                }
                quantity {
                  quantity
                }
              }
              ... on DiscountPercentage {
                percentage
              }
            }
          }
          customerSelection {
            __typename
            ... on DiscountCustomerAll {
              allCustomers
            }
            ... on DiscountCustomerSegments {
              segments {
                creationDate
                id
                lastEditDate
                name
                query
              }
            }
            ... on DiscountCustomers {
              customers {
                id
              }
            }
          }
          discountClass
          endsAt
          hasTimelineComment
          shareableUrls {
            url
            title
          }
          startsAt
          status
          summary
          title
          totalSales {
            amount
            currencyCode
          }
          updatedAt
          usageLimit
          usesPerOrderLimit
        }
        ... on DiscountCodeFreeShipping {
          appliesOncePerCustomer
          appliesOnOneTimePurchase
          appliesOnSubscription
          asyncUsageCount
          codesFirst250: codes(first: 250) {
            nodes {
              id
              code
            }
          }
          codesCount {
            count
            precision
          }
          combinesWith {
            orderDiscounts
            productDiscounts
            shippingDiscounts
          }
          createdAt
          customerSelection {
            __typename
            ... on DiscountCustomerAll {
              allCustomers
            }
            ... on DiscountCustomerSegments {
              segments {
                creationDate
                id
                lastEditDate
                name
                query
              }
            }
            ... on DiscountCustomers {
              customers {
                id
              }
            }
          }
          destinationSelection {
            __typename
            ... on DiscountCountries {
              countries
              includeRestOfWorld
            }
            ... on DiscountCountryAll {
              allCountries
            }
          }
          discountClass
          endsAt
          hasTimelineComment
          maximumShippingPrice {
            amount
            currencyCode
          }
          minimumRequirement {
            __typename
            ... on DiscountMinimumQuantity {
              greaterThanOrEqualToQuantity
            }
            ... on DiscountMinimumSubtotal {
              greaterThanOrEqualToSubtotal {
                amount
                currencyCode
              }
            }
          }
          recurringCycleLimit
          shareableUrls {
            targetItemImage {
              id
              url
            }
            targetType
            title
            url
          }
          shortSummary
          startsAt
          status
          summary
          title
          totalSales {
            amount
            currencyCode
          }
          updatedAt
          usageLimit
        }
        ... on DiscountAutomaticApp {
          appDiscountType {
            app {
              apiKey
            }
            appBridge {
              createPath
              detailsPath
            }
            appKey
            description
            discountClass
            functionId
            targetType
            title
          }
          asyncUsageCount
          combinesWith {
            orderDiscounts
            productDiscounts
            shippingDiscounts
          }
          createdAt
          discountClass
          discountId
          startsAt
          endsAt
          errorHistory {
            errorsFirstOccurredAt
            firstOccurredAt
            hasBeenSharedSinceLastError
            hasSharedRecentErrors
          }
          status
          title
          updatedAt
        }
        ... on DiscountAutomaticBasic {
          asyncUsageCount
          combinesWith {
            orderDiscounts
            productDiscounts
            shippingDiscounts
          }
          createdAt
          customerGets {
            appliesOnOneTimePurchase
            appliesOnSubscription
            items {
              __typename
              ... on AllDiscountItems {
                allItems
              }
              ... on DiscountCollections {
                collectionsFirst250: collections(first: 250) {
                  nodes {
                    id
                  }
                }
              }
              ... on DiscountProducts {
                productsFirst250: products(first: 250) {
                  nodes {
                    id
                  }
                }
                productVariantsFirst250: productVariants(first: 250) {
                  nodes {
                    id
                  }
                }
              }
            }
            value {
              __typename
              ... on DiscountAmount {
                amount {
                  amount
                  currencyCode
                }
                appliesOnEachItem
              }
              ... on DiscountOnQuantity {
                effect {
                  __typename
                  ... on DiscountAmount {
                    amount {
                      amount
                      currencyCode
                    }
                    appliesOnEachItem
                  }
                  ... on DiscountPercentage {
                    percentage
                  }
                }
                quantity {
                  quantity
                }
              }
              ... on DiscountPercentage {
                percentage
              }
            }
          }
          discountClass
          endsAt
          minimumRequirement {
            __typename
            ... on DiscountMinimumQuantity {
              greaterThanOrEqualToQuantity
            }
            ... on DiscountMinimumSubtotal {
              greaterThanOrEqualToSubtotal {
                amount
                currencyCode
              }
            }
          }
          recurringCycleLimit
          shortSummary
          startsAt
          status
          summary
          title
          updatedAt
        }
        ... on DiscountAutomaticBxgy {
          asyncUsageCount
          combinesWith {
            orderDiscounts
            productDiscounts
            shippingDiscounts
          }
          createdAt
          customerBuys {
            items {
              __typename
              ... on AllDiscountItems {
                allItems
              }
              ... on DiscountCollections {
                collectionsFirst250: collections(first: 250) {
                  nodes {
                    id
                  }
                }
              }
              ... on DiscountProducts {
                productsFirst250: products(first: 250) {
                  nodes {
                    id
                  }
                }
                productVariantsFirst250: productVariants(first: 250) {
                  nodes {
                    id
                  }
                }
              }
            }
            value {
              __typename
              ... on DiscountPurchaseAmount {
                amount
              }
              ... on DiscountQuantity {
                quantity
              }
            }
          }
          customerGets {
            appliesOnOneTimePurchase
            appliesOnSubscription
            items {
              __typename
              ... on AllDiscountItems {
                allItems
              }
              ... on DiscountCollections {
                collectionsFirst250: collections(first: 250) {
                  nodes {
                    id
                  }
                }
              }
              ... on DiscountProducts {
                productsFirst250: products(first: 250) {
                  nodes {
                    id
                  }
                }
                productVariantsFirst250: productVariants(first: 250) {
                  nodes {
                    id
                  }
                }
              }
            }
            value {
              __typename
              ... on DiscountAmount {
                amount {
                  amount
                  currencyCode
                }
                appliesOnEachItem
              }
              ... on DiscountOnQuantity {
                effect {
                  __typename
                  ... on DiscountAmount {
                    amount {
                      amount
                      currencyCode
                    }
                    appliesOnEachItem
                  }
                  ... on DiscountPercentage {
                    percentage
                  }
                }
                quantity {
                  quantity
                }
              }
              ... on DiscountPercentage {
                percentage
              }
            }
          }
          discountClass
          endsAt
          startsAt
          status
          summary
          title
          updatedAt
          usesPerOrderLimit
        }
        ... on DiscountAutomaticFreeShipping {
          appliesOnOneTimePurchase
          appliesOnSubscription
          asyncUsageCount
          combinesWith {
            orderDiscounts
            productDiscounts
            shippingDiscounts
          }
          createdAt
          destinationSelection
          discountClass
          endsAt
          hasTimelineComment
          maximumShippingPrice {
            amount
            currencyCode
          }
          minimumRequirement {
            __typename
            ... on DiscountMinimumQuantity {
              greaterThanOrEqualToQuantity
            }
            ... on DiscountMinimumSubtotal {
              greaterThanOrEqualToSubtotal {
                amount
                currencyCode
              }
            }
          }
          recurringCycleLimit
          shortSummary
          startsAt
          status
          summary
          title
          totalSales {
            amount
            currencyCode
          }
          updatedAt
        }
      }
      metafieldsFirst250: metafields(first: 250) {
        nodes {
          id
          key
          value
        }
      }
    }
    pageInfo {
        endCursor
        hasNextPage
        hasPreviousPage
        startCursor
    }
  }
}
"""

        yield from client.get_graphql_pages(
            query,
            data_items_path="data.discountNodes.nodes[*]",
            pagination_cursor_path="data.discountNodes.pageInfo.endCursor",
            pagination_variable_name="after",
            variables={
                "first": items_per_page,
            },
        )

    @dlt.resource(primary_key="id", write_disposition="merge")
    def taxonomy(items_per_page: int = items_per_page) -> Iterable[TDataItem]:
        client = ShopifyGraphQLApi(
            base_url=shop_url,
            access_token=private_app_password,
            api_version="2024-07",
        )

        query = """
{
  taxonomy {
    categories(first: 250) {
      nodes {
        id
        isArchived
        isLeaf
        isRoot
        level
        name
        parentId
        fullName
        ancestorIds
        attributes(first: 250) {
          nodes {
            ... on TaxonomyAttribute {
              id
            }
            ... on TaxonomyChoiceListAttribute {
              id
              name
            }
            ... on TaxonomyMeasurementAttribute {
              id
              name
            }
          }
        }
      }
      pageInfo {
        endCursor
        hasNextPage
        hasPreviousPage
        startCursor
      }
    }
  }
}
"""

        yield from client.get_graphql_pages(
            query,
            data_items_path="data.taxonomy.categories.nodes[*]",
            pagination_cursor_path="data.taxonomy.categories.pageInfo.endCursor",
            pagination_cursor_has_next_page_path="data.taxonomy.categories.pageInfo.hasNextPage",
            pagination_variable_name="after",
            variables={
                "first": items_per_page,
            },
        )

    @dlt.resource(
        primary_key="id",
        write_disposition="merge",
        columns={
            "id": {
                "data_type": "text",
                "nullable": False,
                "primary_key": True,
                "description": "A globally unique ID for the product.",
            },
            "availablePublicationsCount": {
                "data_type": "json",
                "nullable": False,
                "description": "The number of publications that a resource is published to",
            },
            "category": {
                "data_type": "json",
                "nullable": True,
                "description": "The category of the product from Shopify's Standard Product Taxonomy.",
            },
            "combinedListing": {
                "data_type": "json",
                "nullable": True,
                "description": "A special product type that combines separate products into a single product listing.",
            },
            "combinedListingRole": {
                "data_type": "json",
                "nullable": True,
                "description": "The role of the product in a combined listing.",
            },
            "compareAtPriceRange": {
                "data_type": "json",
                "nullable": True,
                "description": "The compare-at price range of the product in the shop's default currency.",
            },
            "createdAt": {
                "data_type": "timestamp",
                "nullable": False,
                "description": "The date and time when the product was created.",
            },
            "defaultCursor": {
                "data_type": "text",
                "nullable": False,
                "description": "A default cursor that returns the next record sorted by ID.",
            },
            "description": {
                "data_type": "text",
                "nullable": False,
                "description": "A single-line description of the product, with HTML tags removed.",
            },
            "descriptionHtml": {
                "data_type": "text",
                "nullable": False,
                "description": "The description of the product, with HTML tags.",
            },
            "handle": {
                "data_type": "text",
                "nullable": False,
                "description": "A unique, human-readable string of the product's title.",
            },
            "metafields": {
                "data_type": "json",
                "nullable": True,
                "description": "A list of custom fields associated with the product.",
            },
            "options": {
                "data_type": "json",
                "nullable": True,
                "description": "A list of product options, e.g., size, color.",
            },
            "priceRangeV2": {
                "data_type": "json",
                "nullable": False,
                "description": "The minimum and maximum prices of a product.",
            },
            "productType": {
                "data_type": "text",
                "nullable": False,
                "description": "The product type defined by the merchant.",
            },
            "publishedAt": {
                "data_type": "timestamp",
                "nullable": True,
                "description": "The date and time when the product was published.",
            },
            "requiresSellingPlan": {
                "data_type": "bool",
                "nullable": True,
                "description": "Whether the product can only be purchased with a selling plan.",
            },
            "status": {
                "data_type": "text",
                "nullable": False,
                "description": "The product status, which controls visibility across all sales channels.",
            },
            "tags": {
                "data_type": "text",
                "nullable": True,
                "description": "A comma-separated list of searchable keywords associated with the product.",
            },
            "templateSuffix": {
                "data_type": "text",
                "nullable": True,
                "description": "The theme template used when customers view the product in a store.",
            },
            "title": {
                "data_type": "text",
                "nullable": False,
                "description": "The name for the product that displays to customers.",
            },
            "totalInventory": {
                "data_type": "bigint",
                "nullable": False,
                "description": "The quantity of inventory that's in stock.",
            },
            "tracksInventory": {
                "data_type": "bool",
                "nullable": False,
                "description": "Whether inventory tracking is enabled for the product.",
            },
            "updatedAt": {
                "data_type": "timestamp",
                "nullable": False,
                "description": "The date and time when the product was last modified.",
            },
            "variantsFirst250": {
                "data_type": "json",
                "nullable": False,
                "description": "A list of variants associated with the product, first 250.",
            },
            "variantsCount": {
                "data_type": "json",
                "nullable": False,
                "description": "The number of variants associated with the product.",
            },
            "vendor": {
                "data_type": "text",
                "nullable": False,
                "description": "The name of the product's vendor.",
            },
        },
    )
    def products(
        updated_at: dlt.sources.incremental[
            pendulum.DateTime
        ] = dlt.sources.incremental(
            "updatedAt",
            initial_value=start_date_obj,
            end_value=end_date_obj,
            range_end="closed",
            range_start="closed",
        ),
        items_per_page: int = items_per_page,
    ) -> Iterable[TDataItem]:
        client = ShopifyGraphQLApi(
            base_url=shop_url,
            access_token=private_app_password,
            api_version="2024-07",
        )

        query = """
query products($after: String, $query: String, $first: Int)  {
  products(after: $after, first: $first, query: $query) {
    nodes {
      availablePublicationsCount {
        count
        precision
      }
      category {
        id
      }
      combinedListing {
        parentProduct {
          id
        }
      }
      combinedListingRole
      compareAtPriceRange {
        maxVariantCompareAtPrice {
          amount
          currencyCode
        }
        minVariantCompareAtPrice {
          amount
          currencyCode
        }
      }
      createdAt
      defaultCursor
      description
      descriptionHtml
      handle
      id
      metafields(first: 250) {
        nodes {
          id
          key
          value
        }
      }
      options {
        linkedMetafield {
          key
          namespace
        }
        name
        optionValues {
          hasVariants
          id
          linkedMetafieldValue
          name
        }
        values
        id
        position
      }
      priceRangeV2 {
        maxVariantPrice {
          amount
          currencyCode
        }
        minVariantPrice {
          amount
          currencyCode
        }
      }
      productType
      publishedAt
      requiresSellingPlan
      status
      tags
      templateSuffix
      totalInventory
      title
      tracksInventory
      updatedAt
      vendor
      variantsCount {
        count
        precision
      }
      variantsFirst250: variants(first: 250) {
        nodes {
          id
          sku
        }
      }
    }
    pageInfo {
      endCursor
      hasNextPage
      hasPreviousPage
    }
  }
}
"""
        variables = {
            "first": items_per_page,
            "query": f"updated_at:>'{updated_at.last_value.isoformat()}'",
        }

        yield from client.get_graphql_pages(
            query,
            data_items_path="data.products.nodes[*]",
            pagination_cursor_path="data.products.pageInfo.endCursor",
            pagination_cursor_has_next_page_path="data.products.pageInfo.hasNextPage",
            pagination_variable_name="after",
            variables=variables,
        )

    return (
        products,
        products_legacy,
        orders,
        customers,
        inventory_items,
        transactions,
        balance,
        events,
        price_rules,
        discounts,
        taxonomy,
    )



from typing import Iterator, List, Optional, Tuple

from dlt.common.typing import DictStrAny, StrAny
from dlt.common.utils import chunks
from dlt.sources.helpers import requests

from .queries import COMMENT_REACTIONS_QUERY, ISSUES_QUERY, RATE_LIMIT, STARGAZERS_QUERY
from .settings import GRAPHQL_API_BASE_URL, REST_API_BASE_URL


#
# Shared
#
def _get_auth_header(access_token: Optional[str]) -> StrAny:
    if access_token:
        return {"Authorization": f"Bearer {access_token}"}
    else:
        # REST API works without access token (with high rate limits)
        return {}


#
# Rest API helpers
#
def get_rest_pages(access_token: Optional[str], query: str) -> Iterator[List[StrAny]]:
    def _request(page_url: str) -> requests.Response:
        r = requests.get(page_url, headers=_get_auth_header(access_token))
        print(
            f"got page {page_url}, requests left: " + r.headers["x-ratelimit-remaining"]
        )
        return r

    next_page_url = REST_API_BASE_URL + query
    while True:
        r: requests.Response = _request(next_page_url)
        page_items = r.json()
        if len(page_items) == 0:
            break
        yield page_items
        if "next" not in r.links:
            break
        next_page_url = r.links["next"]["url"]


#
# GraphQL API helpers
#
def get_stargazers(
    owner: str,
    name: str,
    access_token: str,
    items_per_page: int,
    max_items: Optional[int],
) -> Iterator[Iterator[StrAny]]:
    variables = {"owner": owner, "name": name, "items_per_page": items_per_page}
    for page_items in _get_graphql_pages(
        access_token, STARGAZERS_QUERY, variables, "stargazers", max_items
    ):
        yield map(
            lambda item: {"starredAt": item["starredAt"], "user": item["node"]},
            page_items,
        )


def get_reactions_data(
    node_type: str,
    owner: str,
    name: str,
    access_token: str,
    items_per_page: int,
    max_items: Optional[int],
) -> Iterator[Iterator[StrAny]]:
    variables = {
        "owner": owner,
        "name": name,
        "issues_per_page": items_per_page,
        "first_reactions": 100,
        "first_comments": 100,
        "node_type": node_type,
    }
    for page_items in _get_graphql_pages(
        access_token, ISSUES_QUERY % node_type, variables, node_type, max_items
    ):
        # use reactionGroups to query for reactions to comments that have any reactions. reduces cost by 10-50x
        reacted_comment_ids = {}
        for item in page_items:
            for comment in item["comments"]["nodes"]:
                if any(group["createdAt"] for group in comment["reactionGroups"]):
                    # print(f"for comment {comment['id']}: has reaction")
                    reacted_comment_ids[comment["id"]] = comment
                # if "reactionGroups" in comment:
                comment.pop("reactionGroups", None)

        # get comment reactions by querying comment nodes separately
        comment_reactions = _get_comment_reaction(
            list(reacted_comment_ids.keys()), access_token
        )
        # attach the reaction nodes where they should be
        for comment in comment_reactions.values():
            comment_id = comment["id"]
            reacted_comment_ids[comment_id]["reactions"] = comment["reactions"]
        yield map(_extract_nested_nodes, page_items)


def _extract_top_connection(data: StrAny, node_type: str) -> StrAny:
    assert (
        isinstance(data, dict) and len(data) == 1
    ), f"The data with list of {node_type} must be a dictionary and contain only one element"
    data = next(iter(data.values()))
    return data[node_type]  # type: ignore


def _extract_nested_nodes(item: DictStrAny) -> DictStrAny:
    """Recursively moves `nodes` and `totalCount` to reduce nesting."""
    item["reactions_totalCount"] = item["reactions"].get("totalCount", 0)
    item["reactions"] = item["reactions"]["nodes"]
    comments = item["comments"]
    item["comments_totalCount"] = item["comments"].get("totalCount", 0)
    for comment in comments["nodes"]:
        if "reactions" in comment:
            comment["reactions_totalCount"] = comment["reactions"].get("totalCount", 0)
            comment["reactions"] = comment["reactions"]["nodes"]
    item["comments"] = comments["nodes"]
    return item


def _run_graphql_query(
    access_token: str, query: str, variables: DictStrAny
) -> Tuple[StrAny, StrAny]:
    def _request() -> requests.Response:
        r = requests.post(
            GRAPHQL_API_BASE_URL,
            json={"query": query, "variables": variables},
            headers=_get_auth_header(access_token),
        )
        return r

    data = _request().json()
    if "errors" in data:
        raise ValueError(data)
    data = data["data"]
    # pop rate limits
    rate_limit = data.pop("rateLimit", {"cost": 0, "remaining": 0})
    return data, rate_limit


def _get_graphql_pages(
    access_token: str, query: str, variables: DictStrAny, node_type: str, max_items: int
) -> Iterator[List[DictStrAny]]:
    items_count = 0
    while True:
        data, rate_limit = _run_graphql_query(access_token, query, variables)
        top_connection = _extract_top_connection(data, node_type)
        data_items = (
            top_connection["nodes"]
            if "nodes" in top_connection
            else top_connection["edges"]
        )
        items_count += len(data_items)
        print(
            f'Got {len(data_items)}/{items_count} {node_type}s, query cost {rate_limit["cost"]}, remaining credits: {rate_limit["remaining"]}'
        )
        if data_items:
            yield data_items
        else:
            return
        # print(data["repository"][node_type]["pageInfo"]["endCursor"])
        variables["page_after"] = _extract_top_connection(data, node_type)["pageInfo"][
            "endCursor"
        ]
        if max_items and items_count >= max_items:
            print(f"Max items limit reached: {items_count} >= {max_items}")
            return


def _get_comment_reaction(comment_ids: List[str], access_token: str) -> StrAny:
    """Builds a query from a list of comment nodes and returns associated reactions."""
    idx = 0
    data: DictStrAny = {}
    for page_chunk in chunks(comment_ids, 50):
        subs = []
        for comment_id in page_chunk:
            subs.append(COMMENT_REACTIONS_QUERY % (idx, comment_id))
            idx += 1
        subs.append(RATE_LIMIT)
        query = "{" + ",\n".join(subs) + "}"
        # print(query)
        page, rate_limit = _run_graphql_query(access_token, query, {})
        print(
            f'Got {len(page)} comments, query cost {rate_limit["cost"]}, remaining credits: {rate_limit["remaining"]}'
        )
        data.update(page)
    return data
